import os, sys, multiprocessing
from fractions import Fraction
from typing import Union
import numpy as np
from .kshell_exceptions import DataStructureNotAccountedForError
from .general_utilities import level_plot, level_density

atomic_numbers = {
    "oxygen": 8, "fluorine": 9, "neon": 10, "sodium": 11, "magnesium": 12,
    "aluminium": 13, "silicon": 14, "phosphorus": 15, "sulfur": 16,
    "chlorine": 17, "argon": 18
}

atomic_numbers_reversed = {
    8: 'oxygen', 9: 'fluorine', 10: 'neon', 11: 'sodium', 12: 'magnesium',
    13: 'aluminium', 14: 'silicon', 15: 'phosphorus', 16: 'sulfur',
    17: 'chlorine', 18: 'argon'
}

def generate_states(
    start: int = 0,
    stop: int = 14,
    n_states: int = 100,
    parity: Union[str, int] = "both"
    ):
    """
    Generate correct string for input to kshell_ui.py when asked for
    which states to calculate.

    Examples
    --------
    For input start = 0, stop = 14, n_states = 200:
    0+200, 0.5+200, 1+200, 1.5+200, 2+200, 2.5+200, 3+200, 3.5+200, 4+200,
    4.5+200, 5+200, 5.5+200, 6+200, 6.5+200, 7+200, 7.5+200, 8+200, 8.5+200,
    9+200, 9.5+200, 10+200, 10.5+200, 11+200, 11.5+200, 12+200, 12.5+200,
    13+200, 13.5+200, 14+200, 14.5+200, 0-200, 0.5-200, 1-200, 1.5-200, 2-200,
    2.5-200, 3-200, 3.5-200, 4-200, 4.5-200, 5-200, 5.5-200, 6-200, 6.5-200,
    7-200, 7.5-200, 8-200, 8.5-200, 9-200, 9.5-200, 10-200, 10.5-200, 11-200,
    11.5-200, 12-200, 12.5-200, 13-200, 13.5-200, 14-200, 14.5-200,
    """
    allowed_positive_parity_inputs = ["positive", "pos", "+", "1", "+1", 1, "both"]
    allowed_negative_parity_inputs = ["negative", "neg", "-", "-1", -1, "both"]
    
    def correct_syntax(lst):
        for elem in lst:
            print(elem, end=", ")
    
    if parity in allowed_positive_parity_inputs:
        positive = [f"{i:g}{'+'}{n_states}" for i in np.arange(start, stop+0.5, 0.5)]
        correct_syntax(positive)
    
    if parity in allowed_negative_parity_inputs:
        negative = [f"{i:g}{'-'}{n_states}" for i in np.arange(start, stop+0.5, 0.5)]
        correct_syntax(negative)

class ReadKshellOutput:
    """
    Implemented as class just to avoid returning a bunch of values.
    Access instance attributes instead.
    TODO: Implement call to level_plot as a method.
    TODO: Why is initial parity twice in self.transitions? Fix.
    TODO: Skip -1 spin states. These are bad outputs from KSHELL.

    Attributes
    ----------
    self.Ex:
        Array of excitation energies zeroed at the ground state energy.

    self.BM1:
        Array of [[E, B_decay_prob, E_gamma], ...].
        Reduced transition probabilities for M1.

    self.BE2:
        Array of [[E, B_decay_prob, E_gamma], ...].
        Reduced transition probabilities for E2.

    self.levels:
        Array containing energy, spin, and parity for each excited
        state. [[E, 2*spin, parity], ...].

    self.transitions:
        Mx8 array containing [2*spin_final, parity_initial, Ex_final,
        2*spin_initial, parity_initial, Ex_initial, E_gamma, B(.., i->f)]
    """
    def __init__(self, path: str, load_and_save_to_file: bool):
        """
        TODO: Not necessarry to call all _extract_... methods when
        a directory path is passed.

        Parameters
        ----------
        path : string
            Path of KSHELL output file directory, or path to a specific
            KSHELL data file.

        load_and_save_to_file:
            Toggle saving data as .npy files on / off. If 'overwrite',
            saved .npy files are overwritten.
        """

        self.path = path
        self.load_and_save_to_file = load_and_save_to_file
        # Some attributes might not be altered, depending on the input file.
        self.fname_summary = None
        self.fname_ptn = None
        self.nucleus = None
        self.model_space = None
        self.proton_partition = None
        self.neutron_partition = None
        self.Ex = None
        self.BM1 = None
        self.BE2 = None
        self.levels = None
        self.transitions = None
        self.transitions_BM1 = None
        self.transitions_BE2 = None
        self.truncation = None
        # Debug.
        self.minus_one_spin_counts = np.array([0, 0])  # The number of skipped -1 spin states for [levels, transitions].

        if isinstance(self.load_and_save_to_file, str) and (self.load_and_save_to_file != "overwrite"):
            msg = "Allowed values for 'load_and_save_to_file' are: 'True', 'False', 'overwrite'."
            msg += f" Got '{self.load_and_save_to_file}'."
            raise ValueError(msg)

        if os.path.isdir(path):
            """
            If input 'path' is a directory containing KSHELL files,
            extract info from both summary and .ptn file.
            """
            for elem in os.listdir(path):
                if elem.startswith("summary"):
                    self.fname_summary = f"{path}/{elem}"
                    self._extract_info_from_summary_fname()
                    self._read_summary()

                elif elem.endswith(".ptn"):
                    self.fname_ptn = f"{path}/{elem}"
                    self._extract_info_from_ptn_fname()
                    self.read_ptn()

        else:
            """
            'path' is a single file, not a directory.
            """
            fname = path.split("/")[-1]

            if fname.startswith("summary"):
                self.fname_summary = path
                self._extract_info_from_summary_fname()
                self._read_summary()

            elif fname.endswith(".ptn"):
                self.fname_ptn = path
                self._extract_info_from_ptn_fname()
                self._read_ptn()

            else:
                msg = f"Handling for file {fname} is not implemented."
                raise DataStructureNotAccountedForError(msg)

    def _extract_info_from_ptn_fname(self):
        """
        Extract nucleus and model space name.
        """
        fname_split = self.fname_ptn.split("/")[-1]
        fname_split = fname_split.split("_")
        self.nucleus = fname_split[0]
        self.model_space = fname_split[1]

    def _read_ptn(self):
        """
        Read KSHELL partition file (.ptn) and extract proton partition,
        neutron partition, and particle-hole truncation data. Save as
        instance attributes.
        """

        line_number = 0
        line_number_inner = 0
        self.truncation = []

        with open(self.fname_ptn, "r") as infile:
            for line in infile:
                line_number += 1
                
                if line.startswith("# proton partition"):
                    for line_inner in infile:
                        """
                        Read until next '#'.
                        """
                        line_number_inner += 1
                        if line_inner.startswith("#"):
                            line = line_inner
                            break
                    
                    self.proton_partition = np.loadtxt(
                        fname = self.fname_ptn,
                        skiprows = line_number,
                        max_rows = line_number_inner
                    )
                    line_number += line_number_inner
                    line_number_inner = 0
                
                if line.startswith("# neutron partition"):
                    for line_inner in infile:
                        """
                        Read until next '#'.
                        """
                        line_number_inner += 1
                        if line_inner.startswith("#"):
                            line = line_inner
                            break
                    
                    self.neutron_partition = np.loadtxt(
                        fname = self.fname_ptn,
                        skiprows = line_number,
                        max_rows = line_number_inner
                    )
                    line_number += line_number_inner
                    line_number_inner = 0

                if line.startswith("# particle-hole truncation"):
                    for line_inner in infile:
                        """
                        Loop over all particle-hole truncation lines.
                        """
                        line_number += 1
                        line_inner_split = line_inner.split()

                        if (len(line_inner_split) < 2):
                            """
                            Condition will probably not get fulfilled.
                            Safety precaution due to indexing in this
                            loop.
                            """
                            break

                        if (line_inner_split[1]).startswith("["):
                            """
                            '[' indicates that 'line_inner' is still
                            containing truncation information.
                            """
                            for colon_index, elem in enumerate(line_inner_split):
                                """
                                Find the index of the colon ':' to
                                decide the orbit numbers and occupation
                                numbers.
                                """
                                if (elem == ":"): break

                            occupation = [int(occ) for occ in line_inner_split[colon_index + 1:]]   # [min, max].
                            orbit_numbers = "".join(line_inner_split[1:colon_index])
                            orbit_numbers = orbit_numbers.replace("[", "")
                            orbit_numbers = orbit_numbers.replace("]", "")
                            orbit_numbers = orbit_numbers.replace(" ", "")  # This can prob. be removed because of the earlier split.
                            orbit_numbers = orbit_numbers.split(",")
                            orbit_numbers = [int(orbit) for orbit in orbit_numbers]
                            
                            for orbit in orbit_numbers:
                                self.truncation.append((orbit, occupation))
                        
                        else:
                            """
                            Line does not contain '[' and thus does not
                            contain truncation information.
                            """
                            break

    def _extract_info_from_summary_fname(self):
        """
        Extract nucleus and model space name.
        """
        fname_split = self.fname_summary.split("/")[-1]  # Remove path.
        fname_split = fname_split.split("_")
        self.nucleus = fname_split[1]
        self.model_space = fname_split[2][:-4]  # Remove .txt and keep model space name.

    def _read_summary(self):
        """
        Read energy level data, transition probabilities and transition
        strengths from KSHELL output files.

        TODO: Change all the substring indexing to something more
        rigorous, like string.split and similar.

        Raises
        ------
        DataStructureNotAccountedForError
            If the KSHELL file has unexpected structure / syntax.
        """
        npy_path = "tmp"
        base_fname = self.path.split("/")[-1][:-4]

        try:
            os.mkdir(npy_path)
        except FileExistsError:
            pass
            
        levels_fname = f"{npy_path}/{base_fname}_levels.npy"
        transitions_fname = f"{npy_path}/{base_fname}_transitions.npy"
        transitions_BM1_fname = f"{npy_path}/{base_fname}_transitions_BM1.npy"
        transitions_BE2_fname = f"{npy_path}/{base_fname}_transitions_BE2.npy"
        Ex_fname = f"{npy_path}/{base_fname}_Ex.npy"
        BM1_fname = f"{npy_path}/{base_fname}_BM1.npy"
        BE2_fname = f"{npy_path}/{base_fname}_BE2.npy"
        debug_fname = f"{npy_path}/{base_fname}_debug.npy"

        fnames = [
            levels_fname, transitions_fname, Ex_fname, BM1_fname, BE2_fname,
            transitions_BM1_fname, transitions_BE2_fname, debug_fname
        ]

        if self.load_and_save_to_file != "overwrite":
            """
            Do not load files if overwrite parameter has been passed.
            """
            if all([os.path.isfile(fname) for fname in fnames]) and self.load_and_save_to_file:
                """
                If all files exist, load them. If any of the files does
                not exist, all will be generated.
                """
                self.Ex = np.load(file=Ex_fname, allow_pickle=True)
                self.BM1 = np.load(file=BM1_fname, allow_pickle=True)
                self.BE2 = np.load(file=BE2_fname, allow_pickle=True)
                self.levels = np.load(file=levels_fname, allow_pickle=True)
                self.transitions = np.load(file=transitions_fname, allow_pickle=True)
                self.transitions_BM1 = np.load(file=transitions_BM1_fname, allow_pickle=True)
                self.transitions_BE2 = np.load(file=transitions_BE2_fname, allow_pickle=True)
                self.debug = np.load(file=debug_fname, allow_pickle=True)
                print("Summary data loaded from .npy!")
                return

        def load_energy_levels(infile):
            for _ in range(3): infile.readline()
            for line in infile:
                try:
                    tmp = line.split()
                    
                    if tmp[1] == "-1":
                        """
                        -1 spin states in the KSHELL data file indicates
                        bad states which should not be included.
                        """
                        self.minus_one_spin_counts[0] += 1  # Debug.
                        continue
                    
                    self.Ex.append(float(tmp[6]))
                    parity = 1 if tmp[2] == "+" else -1
                    self.levels.append([float(tmp[5]), 2*float(Fraction(tmp[1])), parity])
                except IndexError:
                    """
                    End of energies.
                    """
                    break

        def load_transition_probabilities(infile, reduced_transition_prob_list):
            """
            Parameters
            ----------
            infile:
                The KSHELL summary file.

            reduced_transition_prob_list:
                List for storing B(M1) or B(E2) values.
            """
            for _ in range(2): infile.readline()
            for line in infile:
                try:
                    """
                    Example of possible lines in file:
                    J_i    Ex_i     J_f    Ex_f   dE        B(M1)->         B(M1)<- 
                    2+(11) 18.393 2+(10) 17.791 0.602 0.1(  0.0) 0.1( 0.0)
                    3/2+( 1) 0.072 5/2+( 1) 0.000 0.071 0.127( 0.07) 0.084( 0.05)
                    2+(10) 17.791 2+( 1) 5.172 12.619 0.006( 0.00) 0.006( 0.00)
                    3+( 8) 19.503 2+(11) 18.393 1.111 0.000( 0.00) 0.000( 0.00)
                    1+( 7) 19.408 2+( 9) 16.111 3.297 0.005( 0.00) 0.003( 0.00)
                    5.0+(60) 32.170  4.0+(100) 31.734  0.436    0.198( 0.11)    0.242( 0.14)
                    0.0+(46)', '47.248', '1.0+(97)', '45.384', '1.864', '23.973(13.39)', '7.991(', '4.46)
                    """
                    tmp = line.split()
                    len_tmp = len(tmp)
                    case = None # Used for identifying which if-else case reads wrong.
                    
                    # Location of initial parity is common for all cases.
                    parity_idx = tmp[0].index("(") - 1 # Find index of initial parity.
                    parity_initial = 1 if tmp[0][parity_idx] == "+" else -1
                    parity_symbol = tmp[0][parity_idx]
                    
                    # Location of initial spin is common for all cases.
                    spin_initial = float(Fraction(tmp[0][:parity_idx]))
                    
                    if (tmp[1][-1] != ")") and (tmp[3][-1] != ")") and (len_tmp == 9):
                        """
                        Example:
                        J_i    Ex_i     J_f    Ex_f   dE        B(M1)->         B(M1)<- 
                        2+(11)   18.393  2+(10)    17.791  0.602    0.1(    0.0)    0.1(    0.0)
                        5.0+(60) 32.170  4.0+(100) 31.734  0.436    0.198( 0.11)    0.242( 0.14)
                        """
                        case = 0
                        E_gamma = float(tmp[4])
                        Ex_initial = float(tmp[1])
                        reduced_transition_prob = float(tmp[5][:-1])    # B(M1) or B(E2).
                        spin_final = float(Fraction(tmp[2].split(parity_symbol)[0]))
                        Ex_final = float(tmp[3])

                    elif (tmp[1][-1] != ")") and (tmp[3][-1] == ")") and (len_tmp == 10):
                        """
                        Example:
                        J_i    Ex_i     J_f    Ex_f   dE        B(M1)->         B(M1)<- 
                        2+(10) 17.791 2+( 1) 5.172 12.619 0.006( 0.00) 0.006( 0.00)
                        """
                        case = 1
                        E_gamma = float(tmp[5])
                        Ex_initial = float(tmp[1])
                        reduced_transition_prob = float(tmp[6][:-1])
                        spin_final = float(Fraction(tmp[2][:-2]))
                        Ex_final = float(tmp[4])
                    
                    elif (tmp[1][-1] == ")") and (tmp[4][-1] != ")") and (len_tmp == 10):
                        """
                        Example:
                        J_i    Ex_i     J_f    Ex_f   dE        B(M1)->         B(M1)<- 
                        3+( 8)   19.503 2+(11)    18.393 1.111 0.000( 0.00) 0.000( 0.00)
                        1.0+( 1) 5.357  0.0+(103) 0.000  5.357 0.002( 0.00) 0.007( 0.00)
                        """
                        case = 2
                        E_gamma = float(tmp[5])
                        Ex_initial = float(tmp[2])
                        reduced_transition_prob = float(tmp[6][:-1])
                        spin_final = float(Fraction(tmp[3].split(parity_symbol)[0]))
                        Ex_final = float(tmp[4])

                    elif (tmp[1][-1] == ")") and (tmp[4][-1] == ")") and (len_tmp == 11):
                        """
                        Example:
                        J_i    Ex_i     J_f    Ex_f   dE        B(M1)->         B(M1)<- 
                        1+( 7) 19.408 2+( 9) 16.111 3.297 0.005( 0.00) 0.003( 0.00)
                        """
                        case = 3
                        E_gamma = float(tmp[6])
                        Ex_initial = float(tmp[2])
                        reduced_transition_prob = float(tmp[7][:-1])
                        spin_final = float(Fraction(tmp[3][:-2]))
                        Ex_final = float(tmp[5])

                    elif (tmp[5][-1] == ")") and (tmp[2][-1] == ")") and (len_tmp == 8):
                        """
                        Example:
                        J_i    Ex_i     J_f    Ex_f   dE        B(M1)->         B(M1)<- 
                        0.0+(46) 47.248  1.0+(97) 45.384  1.864   23.973(13.39)    7.991( 4.46)
                        """
                        case = 4
                        E_gamma = float(tmp[4])
                        Ex_initial = float(tmp[1])
                        reduced_transition_prob = float(tmp[5].split("(")[0])
                        spin_final = float(Fraction(tmp[2].split(parity_symbol)[0]))
                        Ex_final = float(tmp[3])

                    else:
                        msg = "ERROR: Structure not accounted for!"
                        msg += f"\n{line=}"
                        raise DataStructureNotAccountedForError(msg)

                    if (spin_final == -1) or (spin_initial == -1):
                        """
                        -1 spin states in the KSHELL data file indicates
                        bad states which should not be included.
                        """
                        self.minus_one_spin_counts[1] += 1  # Debug.
                        continue
                    
                    reduced_transition_prob_list.append([
                        Ex_initial, reduced_transition_prob, E_gamma
                    ])
                    self.transitions.append([
                        2*spin_final, parity_initial, Ex_final, 2*spin_initial,
                        parity_initial, Ex_initial, E_gamma,
                        reduced_transition_prob
                    ])

                except ValueError as err:
                    """
                    One of the float conversions failed indicating that
                    the structure of the line is not accounted for.
                    """
                    msg = "\n" + err.__str__() + f"\n{case=}" + f"\n{line=}"
                    raise DataStructureNotAccountedForError(msg)

                except IndexError:
                    """
                    End of probabilities.
                    """
                    break

        with open(self.fname_summary, "r") as infile:
            for line in infile:
                tmp = line.split()
                try:
                    if tmp[0] == "Energy":
                        self.Ex = []
                        self.levels = [] # [Ei, 2*spin_initial, parity].
                        load_energy_levels(infile)
                    
                    elif tmp[0] == "B(E2)":
                        self.BE2 = []
                        if self.transitions is None:
                            self.transitions = []
                        load_transition_probabilities(infile, self.BE2)
                    
                    elif tmp[0] == "B(M1)":
                        self.BM1 = []
                        if self.transitions is None:
                            self.transitions = []
                        load_transition_probabilities(infile, self.BM1)
                
                except IndexError:
                    """
                    Skip blank lines.
                    """
                    continue

        self.levels = np.array(self.levels)
        self.transitions = np.array(self.transitions)
        self.Ex = np.array(self.Ex)
        self.BM1 = np.array(self.BM1)
        self.BE2 = np.array(self.BE2)
        self.debug = "DEBUG\n"
        self.debug += f"skipped -1 states in levels: {self.minus_one_spin_counts[0]}\n"
        self.debug += f"skipped -1 states in transitions: {self.minus_one_spin_counts[1]}\n"
        self.debug = np.array(self.debug)

        try:
            self.transitions_BE2 = self.transitions[:len(self.BE2)]
            self.transitions_BM1 = self.transitions[len(self.BE2):]
        except TypeError:
            """
            TypeError: len() of unsized object because self.BE2 = None.
            """
            self.transitions_BE2 = np.array(None)
            self.transitions_BM1 = np.array(None)

        if self.load_and_save_to_file:
            np.save(file=levels_fname, arr=self.levels, allow_pickle=True)
            np.save(file=transitions_fname, arr=self.transitions, allow_pickle=True)
            np.save(file=transitions_BM1_fname, arr=self.transitions_BM1, allow_pickle=True)
            np.save(file=transitions_BE2_fname, arr=self.transitions_BE2, allow_pickle=True)
            np.save(file=Ex_fname, arr=self.Ex, allow_pickle=True)
            np.save(file=BM1_fname, arr=self.BM1, allow_pickle=True)
            np.save(file=BE2_fname, arr=self.BE2, allow_pickle=True)
            np.save(file=debug_fname, arr=self.debug, allow_pickle=True)

    def level_plot(self,
        max_spin_states: int = 1_000,
        filter_spins: Union[None, list] = None
        ):
        """
        Wrapper method to include level plot as an attribute to this
        class. Generate a level plot for a single isotope. Spin on the x
        axis, energy on the y axis.

        Parameters
        ----------        
        max_spin_states:
            The maximum amount of states to plot for each spin. Default
            set to a large number to indicate ≈ no limit.

        filter_spins:
            Which spins to include in the plot. If None, all spins are
            plotted.
        """
        level_plot(
            levels = self.levels,
            max_spin_states = max_spin_states,
            filter_spins = filter_spins
        )

    def level_density_plot(self,
            bin_size: Union[int, float]
        ):
        """
        Wrapper method to include level density plotting as
        an attribute to this class. Generate the level density with the
        input bin size.
        """
        level_density(
            energy_levels = self.levels[:, 0],
            bin_size = bin_size,
            plot = True
        )

    @property
    def help(self):
        """
        Generate a list of instance attributes without magic methods.

        Returns
        -------
        help_list : list
            A list of non-magic instance attributes.
        """
        help_list = []
        for elem in dir(self):
            if not elem.startswith("_"):   # Omit magic and private methods.
                help_list.append(elem)
        
        return help_list

def _process_kshell_output_in_parallel(args):
    """
    Simple wrapper for parallelizing loading of KSHELL files.
    """
    filepath, load_and_save_to_file = args
    print(filepath)
    return ReadKshellOutput(filepath, load_and_save_to_file)

def loadtxt(
    path: str,
    is_directory: bool = False,
    filter_: Union[None, str] = None,
    load_and_save_to_file: Union[bool, str] = True
    ) -> list:
    """
    Wrapper for using ReadKshellOutput class as a function.

    Parameters
    ----------
    path:
        Filename (and path) of KSHELL output data file, or path to
        directory containing sub-directories with KSHELL output data.
    
    is_directory:
        If True, and 'path' is a directory containing sub-directories
        with KSHELL data files, the contents of 'path' will be scanned
        for KSHELL data files. Currently supports only summary files.

    filter_:
        NOTE: Shouldnt the type be list, not str?

    load_and_save_to_file:
        Toggle saving data as .npy files on / off. If 'overwrite', saved
        .npy files are overwritten.

    Returns
    -------
    data : list
        List of instances with data from KSHELL data file as attributes.
    """
    all_fnames = None
    data = []
    if (is_directory) and (not os.path.isdir(path)):
        msg = f"{path} is not a directory"
        raise NotADirectoryError(msg)

    elif (not is_directory) and (not os.path.isfile(path)):
        msg = f"{path} is not a file"
        raise FileNotFoundError(msg)

    elif (is_directory) and (os.path.isdir(path)):
        all_fnames = {}

        for element in sorted(os.listdir(path)):
            """
            List all content in path.
            """
            if os.path.isdir(path + element):
                """
                If element is a directory, enter it to find data files.
                """
                all_fnames[element] = []    # Create blank list entry in dict for current element.
                for isotope in os.listdir(path + element):
                    """
                    List all content in the element directory.
                    """
                    if isotope.startswith("summary") and isotope.endswith(".txt"):
                        """
                        Extract summary data files.
                        """
                        try:
                            """
                            Example: O16.
                            """
                            n_neutrons = int(isotope[9:11])
                        except ValueError:
                            """
                            Example: Ne20.
                            """
                            n_neutrons = int(isotope[10:12])

                        n_neutrons -= atomic_numbers[element.split("_")[1]]
                        all_fnames[element].append([element + "/" + isotope, n_neutrons])
        
        pool = multiprocessing.Pool()
        for key in all_fnames:
            """
            Sort each list in the dict by the number of neutrons. Loop
            over all directories in 'all_fnames' and extract KSHELL data
            and append to a list.
            """
            if filter_ is not None:
                if key.split("_")[1] not in filter_:
                    """
                    Skip elements not in filter_.
                    """
                    continue

            all_fnames[key].sort(key=lambda tup: tup[1])   # Why not do this when directory is listed?
            sub_fnames = all_fnames[key]
            arg_list = [(path + i[0], load_and_save_to_file) for i in sub_fnames]
            data += pool.map(_process_kshell_output_in_parallel, arg_list)

    else:
        """
        Only a single KSHELL data file.
        """
        data.append(ReadKshellOutput(path, load_and_save_to_file))

    if not data:
        msg = "No KSHELL data loaded. Most likely error is that the given"
        msg += f" directory has no KSHELL data files. {path=}"
        raise RuntimeError(msg)

    return data



import os
import numpy as np
import matplotlib.pyplot as plt
from .kshell_utilities import atomic_numbers, loadtxt
from .general_utilities import create_spin_parity_list, strength_function_average

class LEE:
    def __init__(self, directory):
        self.bin_width = 0.2
        self.E_max = 30
        self.Ex_min = 0 # Lower limit for emitted gamma energy [MeV].
        self.Ex_max = 30 # Upper limit for emitted gamma energy [MeV].
        n_bins = int(np.ceil(self.E_max/self.bin_width))
        E_max_adjusted = self.bin_width*n_bins
        bins = np.linspace(0, E_max_adjusted, n_bins + 1)
        self.bins_middle = (bins[0: -1] + bins[1:])/2

        self.all_fnames = {}

        self.directory = directory

        for element in sorted(os.listdir(self.directory)):
            """
            List all content in self.directory.
            """
            if os.path.isdir(f"{self.directory}/{element}"):
                """
                If element is a directory, enter it to find data files.
                """
                self.all_fnames[element] = []    # Create blank entry in dict for current element.
                for isotope in os.listdir(f"{self.directory}/{element}"):
                    """
                    List all content in the element directory.
                    """
                    if isotope.startswith("summary"):
                        """
                        Extract summary data files.
                        """
                        try:
                            """
                            Example: O16.
                            """
                            n_neutrons = int(isotope[9:11])
                        except ValueError:
                            """
                            Example: Ne20.
                            """
                            n_neutrons = int(isotope[10:12])

                        n_neutrons -= atomic_numbers[element.split("_")[1]]
                        
                        self.all_fnames[element].append([f"{element}/{isotope}", n_neutrons])
        
        for key in self.all_fnames:
            """
            Sort each list in the dict by the number of neutrons.
            """
            self.all_fnames[key].sort(key=lambda tup: tup[1])   # Why not do this when directory is listed?

    # def plot_gsf(self, isotope_name):
    #     """
    #     Plot the gamma strength function for a single isotope.

    #     isotope_name : string
    #         Examples: S24, Ne30.

    #     Raises
    #     ------
    #     ValueError
    #         If isotope_name cannot be found in the calculated data
    #         files.
    #     """
    #     fname = None
        
    #     for fnames in self.fnames_combined:
    #         for i in range(len(fnames)):
    #             if isotope_name in fnames[i][0]:
    #                 fname = fnames[i][0]

    #     if fname is None:
    #         msg = f"Isotope name '{isotope_name}' is not a valid name."
    #         raise ValueError(msg)

    #     res = loadtxt(self.directory + fname)

    #     _, ax = plt.subplots()

    #     Jpi_list = create_jpi_list(res.levels[:, 1], None)
    #     E_gs = res.levels[0, 0]
    #     res.transitions[:, 2] += E_gs   # Add ground state energy for compatibility with Jørgen.

    #     gsf = strength_function_average(
    #         levels = res.levels,
    #         transitions = res.transitions,
    #         Jpi_list = Jpi_list,
    #         bin_width = self.bin_width,
    #         Ex_min = self.Ex_min,    # [MeV].
    #         Ex_max = self.Ex_max,    # [MeV].
    #         multipole_type = "M1"
    #     )

    #     bin_slice = self.bins_middle[0:len(gsf)]
    #     ax.plot(bin_slice, gsf, label=fname)
    #     ax.legend()
    #     ax.set_xlabel(r"$E_{\gamma}$ [MeV]")
    #     ax.set_ylabel(r"gsf [MeV$^{-3}$]")
    #     plt.show()


    def calculate_low_energy_enhancement(self, filter=None):
        """
        Recreate the figure from Jørgens article.
        """
        self.labels = []    # Suggested labels for plotting.
        self.ratios = []
        self.n_neutrons = []

        for key in self.all_fnames:
            """
            Loop over all elements (grunnstoff).
            """
            fnames = self.all_fnames[key]   # For compatibility with old code.
            if filter is not None:
                if key.split("_")[1] not in filter:
                    """
                    Skip elements not in filter.
                    """
                    continue
            
            ratios = [] # Reset ratio for every new element.
            for i in range(len(fnames)):
                """
                Loop over all isotopes per element.
                """
                try:
                    res = loadtxt(f"{self.directory}/{fnames[i][0]}")
                except FileNotFoundError:
                    print(f"File {fnames[i][0]} skipped! File not found.")
                    ratios.append(None) # Maintain correct list length for plotting.
                    continue

                Jpi_list = create_spin_parity_list(
                    spins = res.levels[:, 1],
                    parities = res.levels[:, 2]
                )
                E_gs = res.levels[0, 0]

                try:
                    res.transitions[:, 2] += E_gs   # Add ground state energy for compatibility with Jørgen.
                except IndexError:
                    print(f"File {fnames[i][0]} skipped! Too few / no energy levels are present in this data file.")
                    ratios.append(None) # Maintain correct list length for plotting.
                    continue
                
                try:
                    gsf = strength_function_average(
                        levels = res.levels,
                        transitions = res.transitions,
                        Jpi_list = Jpi_list,
                        bin_width = self.bin_width,
                        Ex_min = self.Ex_min,    # [MeV].
                        Ex_max = self.Ex_max,    # [MeV].
                        multipole_type = "M1"
                    )
                except IndexError:
                    print(f"File {fnames[i][0]} skipped! That unknown index out of bounds error in ksutil.")
                    ratios.append(None)
                    continue

                # Sum gsf for low and high energy range and take the ratio.
                bin_slice = self.bins_middle[0:len(gsf)]
                low_idx = (bin_slice <= 2)
                high_idx = (bin_slice <= 6) == (2 <= bin_slice)
                low = np.sum(gsf[low_idx])
                high = np.sum(gsf[high_idx])
                low_high_ratio = low/high
                ratios.append(low_high_ratio)

                print(f"{fnames[i][0]} loaded")

            if all(elem is None for elem in ratios):
                """
                Skip current element if no ratios are calculated.
                """
                continue
            
            self.labels.append(fnames[0][0][:fnames[0][0].index("/")])
            self.n_neutrons.append([n_neutrons for _, n_neutrons in fnames])
            self.ratios.append(ratios)


    def quick_plot(self):
        fig, ax = plt.subplots()
        for i in range(len(self.n_neutrons)):
            ax.plot(self.n_neutrons[i], self.ratios[i], ".--", label=self.labels[i])
            ax.set_yscale("log")
            ax.set_xlabel("N")
            ax.set_ylabel("Rel. amount of low-energy strength")
            ax.legend()

        plt.show()


def low_energy_enhancement(directory):
    """
    Wrapper for easier usage.

    Parameters
    ----------
    directory : string
        Directory containing subfolders with KSHELL data.

    Returns
    -------
    res : kshell_utilities.low_energy_enhancement.LEE
        Class instance containing LEE data.
    """
    res = LEE(directory)
    res.calculate_low_energy_enhancement()

    return res


import sys
import numpy as np 
# import matplotlib.pyplot as plt 

# Scripts to sort levels and transition strengths from KSHELL 
# into Ex-Eg matrix energy bins, and to make level density and 
# gamma strength function.
# Plus other useful functions for shell model stuff.


def div0( a, b ):
    """ division function designed to ignore / 0, i.e. div0( [-1, 0, 1], 0 ) -> [0, 0, 0] """
    with np.errstate(divide='ignore', invalid='ignore'):
        c = np.true_divide( a, b )
        c[ ~ np.isfinite( c )] = 0    # -inf inf NaN
    return c


def read_energy_levels(fname):
    """
    Reads levels from a KSHELL summary file, returns Nx3 matrix of 
    [Ei, 2*Ji, parity], where E is absolute energy of level and parity 1=+, 0=-
    """
    levels = []
    with open(fname, 'r') as f:
        lines = f.readlines()
        i_start = -1
        
        for i in range(len(lines)):
            if len(lines[i].split()) < 1:
                continue
            
            if lines[i].split()[0] == "Energy":
                i_start = i+4
                break
        
        for i in range(i_start, len(lines)):
            if len(lines[i].split()) < 1:
                break
            
            words = lines[i].split()
            num_parity = 1 if words[2] == "+" else -1
            levels.append([float(words[5]), float(words[1]), num_parity])
        
        return np.asarray(levels) 


def read_transition_strengths(fname, multipole_type="M1"):
    """
    Parameters
    ----------
    fname : string
        Filename of KSHELL data file.

    multipole_type : string
        Choose whether to calculate for 'M1' or 'E2'.

    Returns
    -------
    transitions : numpy.ndarray
        Matrix with format [2Jf, pi, Ef, 2Ji, pi, Ei, Ex, B(.., i->f)].
    """
    transitions = []
    
    with open(fname, 'r') as f:
        lines = f.readlines()
        i_start = -1
        for i in range(len(lines)):
            """
            Iterate file until B({multipole_type:s}) is found.
            """
            if len(lines[i].split()) < 1:
                """
                Skip empty lines.
                """
                continue
            
            if lines[i].split()[0] == f"B({multipole_type:s})":
                """
                Correct multipole header found.  Update start index.
                """
                i_start = i + 2
                break

        for i in range(i_start, len(lines)):
            """
            Read the multipole transition data.
            """
            if len(lines[i].split()) < 1:
                """
                Skip empty lines.
                """
                break

            line = lines[i]
            pi_str = line[25:27].strip()
            if pi_str == "+":
                pi = +1
            elif pi_str == "-":
                pi = -1
            else:
                msg = "From function read_transition_strengths: Could not"
                msg += " assign initial parity. Read value: " + pi_str
                raise Exception(msg)
            

            # NOTE: pf is calculated but not used.
            pf_str = line[4:5].strip()
            if pf_str == "+":
                pf = +1
            elif pf_str == "-":
                pf = -1
            else:
                msg = "From function read_transition_strengths: Could not"
                msg += " assign final parity Read value: " + pf_str
                raise Exception(msg)

            transitions.append([
                float(line[22:25]), # 2Jf.
                pi,                 # pf maybe?
                float(line[34:43]), # Ef.
                float(line[0:3]),   # 2Ji.
                pi,                 # Should these two pi's be equal?
                float(line[12:22]), # Ei.
                float(line[43:51]), # Ex.
                float(line[67:83])  # B(..)<-
            ])

        return np.asarray(transitions)


def read_calc_tbme(fname, tbme_template):
    # Reads tbme expectation values from a KSHELL summary file, returns NxM matrix of 
    # the M SPE+TBMEs for each level N
    # The tbme_template file is used to compare the ordering of <k1k2|V|k3k4;J> because
    # KSHELL can give either ordering of the k1k2, k3k4 pairs.
    calc_tbme = []
    with open(fname, 'r') as f:
        lines = f.readlines()
        i_start = -1
        for i in range(len(lines)):
            if len(lines[i].split())<1: continue
            if lines[i].split()[0] == "SPE&TBME":
                i_start = i+1
                break
        # Start looping through all eigenstates:
        i = i_start
        while True:
            if len(lines[i]) >= 30 and lines[i][0:30] == "    N          <E>    log-file":
                # Start reading all calc_tbme for current eigenstate
                i += 1
                words = lines[i].split()
                Nlevel = int(words[0])
                Elevel = float(words[1])
                i += 1
                words = lines[i].split()
                Nspe = int(words[-1])
                i += 1
                spe = np.zeros((Nspe,3)) # Rows of [N, x_i, <O_i>]
                for j in range(Nspe):
                    words = lines[i].split()
                    spe[j,:] = (int(words[0]), float(words[1]), float(words[2]))
                    i += 1
                # i should now be line starting with TBME
                if not lines[i][0:5] == "TBMEs": raise Exception("Error from read_calc_tbme(). Error in summary file format?")
                words = lines[i].split()
                Ntbme = int(words[-1])
                i += 1
                tbme = np.zeros((Ntbme, 9))
                for j in range(Ntbme):
                    words = lines[i].split()
                    k1, k2, k3, k4, jj, iprty, ipn, xi, Oi = int(words[0]), int(words[1]), int(words[2]), int(words[3]), int(words[4]), int(words[5]), int(words[6]), float(words[7]), float(words[8])
                    found_current_tbme = False
                    for i_template in range(Ntbme):
                        l1, l2, l3, l4, jj_t = int(tbme_template[i_template,0]), int(tbme_template[i_template,1]), int(tbme_template[i_template,2]), int(tbme_template[i_template,3]), int(tbme_template[i_template,4])
                        if l1==k1 and l2==k2 and l3==k3 and l4==k4 and jj==jj_t:
                            tbme[j,:] = (k1, k2, k3, k4, jj, iprty, ipn, xi, Oi)
                            found_current_tbme = True
                        elif l1==k3 and l2==k4 and l3==k1 and l4==k2 and jj==jj_t:
                            tbme[j,:] = (k3, k4, k1, k2, jj, iprty, ipn, xi, Oi)
                            found_current_tbme = True
                    if not found_current_tbme: raise Exception("Error in read_calc_tbme(): Did not find matching quantum numbers in tbme_template for TBME no "+str(j))
                    i += 1
                calc_tbme.append({"E":Elevel, "SPE":spe, "TBME":tbme})
            # end reading of one egeinstate
            i += 1
            if i >= len(lines): break

        return calc_tbme


def total_level_density(levels, bin_width, Ex_max):
    # 20170816: This function returns the total level density as a function of Ex.
    Nbins = int(np.ceil(Ex_max/bin_width)) # Make sure the number of bins cover the whole Ex region.
    bin_array = np.linspace(0,bin_width*Nbins,Nbins+1) # Array of lower bin edge energy values
    Egs = levels[0,0]
    rho_total, tmp = np.histogram(levels[:,0]-Egs, bins=bin_array)
    rho_total = rho_total/bin_width # To get density per MeV
    return rho_total


def strength_function_average(
        levels,
        transitions,
        Jpi_list,
        bin_width,
        Ex_min,
        Ex_max,
        multipole_type = "M1"
    ):
    """
    20171009: Updated the way we average over Ex, J, pi to only count pixels with non-zero gSF.
    20170815: This function returns the strength function the way we now think is the correct way:
    By taking only the partial level density corresponding to the specific (Ex, J, pi) pixel in the
    calculation of the strength function, and then averaging over all three variables to produce
    <f(Eg)>.
    This code was first developed in the script strength_function_individual_Jpi.py

    # # Update 20170915: Realized a problem with summing vs averaging, adding this list to fix that
    # # Update 20170920: Realized that the fix was wrong, the original was correct.
    # Ex_already_seen = []
    # for i_Ex in range(Nbins):
    #     Ex_already_seen.append([])

    # 20170920: We thought this was more correct, but now think not.
    # if not Ex in Ex_already_seen[i_Eg]:
    #     B_pixel_count[i_Ex,i_Eg,i_Jpi] += 1
    #     Ex_already_seen[i_Eg].append(Ex)

    Parameters
    ----------
    levels : numpy.ndarray
        Nx3 matrix containing [Ei, 2*Ji, parity] in each row.

    transitions : numpy.ndarray
        Mx8 matrix containing [2Jf, pi, Ef, 2Ji, pi, Ei, Ex, B(.., i->f)]
        in each row.

    Jpi_list : list
        Set a spin window by defining a list of allowed initial
        [[spins, parities], ...]. 

    Ex_min : int, float
        Lower limit for emitted gamma energy [MeV].

    Ex_max : int, float
        Upper limit for emitted gamma energy [MeV].

    multipole_type : string
        Choose whether to calculate for 'M1' or 'E2'.
    """
    Nbins = int(np.ceil(Ex_max/bin_width)) # Make sure the number of bins cover the whole Ex region.
    # bin_array = np.linspace(0, bin_width*Nbins, Nbins + 1) # Array of lower bin edge energy values
    # bin_array_middle = (bin_array[0: -1] + bin_array[1:])/2 # Array of middle bin values
    
    # Find index of first and last bin (lower bin edge) where we put counts.
    # It's important to not include the other Ex bins in the averaging later, because they contain zeros which will pull the average down.
    i_Ex_min = int(np.floor(Ex_min/bin_width)) 
    i_Ex_max = int(np.floor(Ex_max/bin_width))    

    prefactor = {"M1": 11.5473e-9, "E1": 1.047e-6}

    Egs = levels[0, 0] # Read out the absolute ground state energy, so we can get relative energies later.

    # Allocate matrices to store the summed B(M1) values for each pixel,
    # and the number of transitions counted.
    B_pixel_sum = np.zeros((Nbins, Nbins, len(Jpi_list)))
    B_pixel_count = np.zeros((Nbins, Nbins, len(Jpi_list)))

    for i_tr in range(len(transitions[:, 0])):
        """
        Iterate over all transitions in the transitions matrix and put
        in the correct pixel.
        """
        Ex = transitions[i_tr, 2] - Egs # Calculate relative energy.
        if (Ex < Ex_min) or (Ex >= Ex_max):
            """
            Check if transition is within min max limits, skip if not.
            """
            continue

        # Get bin index for Eg and Ex (initial). Indices are defined with respect to the lower bin edge.
        i_Eg = int(np.floor(transitions[i_tr, 6]/bin_width))
        i_Ex = int(np.floor(Ex/bin_width))

        # Read initial spin and parity of level: NOTE: I think the name / index is wrong.
        Ji = int(transitions[i_tr, 0])
        pi = int(transitions[i_tr, 1])
        try:
            """
            Get index for current [Ji, pi] combination in Jpi_list.
            """
            i_Jpi = Jpi_list.index([Ji, pi])
        except ValueError:
            continue

        # Add B(M1) value and increment count to pixel, respectively
        B_pixel_sum[i_Ex, i_Eg, i_Jpi] += transitions[i_tr, 7]
        B_pixel_count[i_Ex, i_Eg, i_Jpi] += 1 # Original.



    # Allocate (Ex, Jpi) matrix to store level density
    rho_ExJpi = np.zeros((Nbins, len(Jpi_list)))
    # Count number of levels for each (Ex, J, pi) pixel.
    for i_l in range(len(levels[:, 0])):
        E, J, pi = levels[i_l]
        
        if (E - Egs) >= Ex_max:
            """
            Skip if level is outside range.
            """
            continue

        i_Ex = int(np.floor((E - Egs)/bin_width))
        try:
            i_Jpi = Jpi_list.index([J, pi])
        except:
            continue
        rho_ExJpi[i_Ex, i_Jpi] += 1

    rho_ExJpi /= bin_width # Normalize to bin width, to get density in MeV^-1.


    # Calculate gamma strength functions for each Ex, J, pi individually, using the partial level density for each J, pi.
    gSF = np.zeros((Nbins, Nbins, len(Jpi_list)))
    a = prefactor[multipole_type] # mu_N^-2 MeV^-2, conversion constant
    for i_Jpi in range(len(Jpi_list)):
        for i_Ex in range(Nbins):
            gSF[i_Ex, :, i_Jpi] = a*rho_ExJpi[i_Ex, i_Jpi]*div0(
                B_pixel_sum[i_Ex, :, i_Jpi],
                B_pixel_count[i_Ex, :, i_Jpi]
            )

    # Return the average gSF(Eg) over all (Ex,J,pi)

    # return gSF[i_Ex_min:i_Ex_max+1,:,:].mean(axis=(0,2))
    # Update 20171009: Took proper care to only average over the non-zero f(Eg,Ex,J,pi) pixels:
    gSF_currentExrange = gSF[i_Ex_min:i_Ex_max + 1, :, :]
    gSF_ExJpiavg = div0(
        gSF_currentExrange.sum(axis = (0, 2)),
        (gSF_currentExrange != 0).sum(axis = (0, 2))
    )
    return gSF_ExJpiavg


def strength_function_average_updated_definition_brute_avg(levels, transitions, Jpi_list, bin_width, Ex_min, Ex_max, type="M1"):
    # 20171009: This is a check to see that the averaging we do over Ex, J, pi really only counts
    # the bins with a non-zero gSF.
    # 20170815: This function returns the strength function the way we now think is the correct way:
    # By taking only the partial level density corresponding to the specific (Ex, J, pi) pixel in the
    # calculation of the strength function, and then averaging over all three variables to produce
    # <f(Eg)>.
    # This code was first developed in the script strength_function_individual_Jpi.py
    Nbins = int(np.ceil(Ex_max/bin_width)) # Make sure the number of bins cover the whole Ex region.
    # print "Ex_max =", Ex_max, "Nbins*bin_width =", Nbins*bin_width
    bin_array = np.linspace(0,bin_width*Nbins,Nbins+1) # Array of lower bin edge energy values
    bin_array_middle = (bin_array[0:-1]+bin_array[1:])/2 # Array of middle bin values
    # Find index of first and last bin (lower bin edge) where we put counts.
    # It's important to not include the other Ex bins in the averaging later, because they contain zeros which will pull the average down.
    i_Ex_min = int(np.floor(Ex_min/bin_width)) 
    i_Ex_max = int(np.floor(Ex_max/bin_width))    

    prefactor = {"M1":    11.5473e-9, "E1": 1.047e-6}

    Egs = levels[0,0] # Read out the absolute ground state energy, so we can get relative energies later

    # Allocate matrices to store the summed B(M1) values for each pixel, and the number of transitions counted
    B_pixel_sum = np.zeros((Nbins,Nbins,len(Jpi_list)))
    B_pixel_count = np.zeros((Nbins,Nbins,len(Jpi_list)))

    # # Update 20170915: Realized a problem with summing vs averaging, adding this list to fix that
    # # Update 20170920: Realized that the fix was wrong, the original was correct.
    # Ex_already_seen = []
    # for i_Ex in range(Nbins):
    #     Ex_already_seen.append([])

    # Loop over all transitions and put in the correct pixel:
    for i_tr in range(len(transitions[:,0])):
        Ex = transitions[i_tr,2] - Egs
        # Check if transition is below Ex_max, skip if not
        if Ex < Ex_min or Ex >= Ex_max:
            continue

        # Get bin index for Eg and Ex (initial). Indices are defined with respect to the lower bin edge.
        i_Eg = int(np.floor(transitions[i_tr,6]/bin_width))
        i_Ex = int(np.floor(Ex/bin_width))

        # Read initial spin and parity of level:
        Ji = int(transitions[i_tr,0])
        pi = int(transitions[i_tr,1])
        # Get index for current [Ji,pi] combination in Jpi_list:
        try:
            i_Jpi = Jpi_list.index([Ji,pi])
        except: 
            continue

        # Add B(M1) value and increment count to pixel, respectively
        B_pixel_sum[i_Ex,i_Eg,i_Jpi] += transitions[i_tr,7]

        # 20170920: We thought this was more correct, but now think not.
        # if not Ex in Ex_already_seen[i_Eg]:
        #     B_pixel_count[i_Ex,i_Eg,i_Jpi] += 1
        #     Ex_already_seen[i_Eg].append(Ex)

        B_pixel_count[i_Ex,i_Eg,i_Jpi] += 1 # Original



    # Allocate (Ex,Jpi) matrix to store level density
    rho_ExJpi = np.zeros((Nbins,len(Jpi_list)))
    # Count number of levels for each (Ex, J, pi) pixel.
    for i_l in range(len(levels[:,0])):
        E, J, pi = levels[i_l]
        # Skip if level is outside range:
        if E-Egs >= Ex_max:
            continue
        i_Ex = int(np.floor((E-Egs)/bin_width))
        try:
            i_Jpi = Jpi_list.index([J,pi])
        except:
            continue
        rho_ExJpi[i_Ex,i_Jpi] += 1

    rho_ExJpi /= bin_width # Normalize to bin width, to get density in MeV^-1


    # Calculate gamma strength functions for each Ex, J, pi individually, using the partial level density for each J, pi.
    gSF = np.zeros((Nbins,Nbins,len(Jpi_list)))
    a = prefactor[type] # mu_N^-2 MeV^-2, conversion constant
    for i_Jpi in range(len(Jpi_list)):
        for i_Ex in range(Nbins):
                                         # a *                     <B(M1; Eg, Ex, J, pi)>                                                                             * rho(Ex, J, pi)    
            gSF[i_Ex,:,i_Jpi] = a * div0(B_pixel_sum[i_Ex,:,i_Jpi], B_pixel_count[i_Ex,:,i_Jpi]) * rho_ExJpi[i_Ex, i_Jpi]
            # TODO if the below smoothing is used: Check the form of the level density. When I tried sigma=3 it looked like the gSF dropped a bit, indicating that the level density was lowered. Maybe smoothe it some other way, such as drawing straight lines?
            # gSF[i_Ex,:,i_Jpi] = a * div0(B_pixel_sum[i_Ex,:,i_Jpi], B_pixel_count[i_Ex,:,i_Jpi]) * smoothe(rho_ExJpi[:, i_Jpi], sigma=1)[i_Ex] # Smoothing the initial level density by a gaussian convolution

    # Return the average gSF(Eg) over all (Ex,J,pi)
    # Test of brute force way, to check:
    gSF_ExJpiavg = np.zeros(Nbins)
    for i_Eg in range(Nbins):
        sum = 0
        counter = 0
        for i_Ex in range(i_Ex_min, i_Ex_max+1):
            for i_Jpi in range(len(Jpi_list)):
                if gSF[i_Ex, i_Eg, i_Jpi] > 0:
                    sum += gSF[i_Ex, i_Eg, i_Jpi]
                    counter += 1
        if counter > 0:
            gSF_ExJpiavg[i_Eg] = sum / counter


    return gSF_ExJpiavg


def strength_function_average_updated_definition_naive_avg(levels, transitions, Jpi_list, bin_width, Ex_min, Ex_max, type="M1"):
    # 20171009: Realized that this was a naive way of averaging over Ex, J and pi.
    # 20170815: This function returns the strength function the way we now think is the correct way:
    # By taking only the partial level density corresponding to the specific (Ex, J, pi) pixel in the
    # calculation of the strength function, and then averaging over all three variables to produce
    # <f(Eg)>.
    # This code was first developed in the script strength_function_individual_Jpi.py
    Nbins = int(np.ceil(Ex_max/bin_width)) # Make sure the number of bins cover the whole Ex region.
    # print "Ex_max =", Ex_max, "Nbins*bin_width =", Nbins*bin_width
    bin_array = np.linspace(0,bin_width*Nbins,Nbins+1) # Array of lower bin edge energy values
    bin_array_middle = (bin_array[0:-1]+bin_array[1:])/2 # Array of middle bin values
    # Find index of first and last bin (lower bin edge) where we put counts.
    # It's important to not include the other Ex bins in the averaging later, because they contain zeros which will pull the average down.
    i_Ex_min = int(np.floor(Ex_min/bin_width)) 
    i_Ex_max = int(np.floor(Ex_max/bin_width))    

    prefactor = {"M1":    11.5473e-9, "E1": 1.047e-6}

    Egs = levels[0,0] # Read out the absolute ground state energy, so we can get relative energies later

    # Allocate matrices to store the summed B(M1) values for each pixel, and the number of transitions counted
    B_pixel_sum = np.zeros((Nbins,Nbins,len(Jpi_list)))
    B_pixel_count = np.zeros((Nbins,Nbins,len(Jpi_list)))

    # # Update 20170915: Realized a problem with summing vs averaging, adding this list to fix that
    # # Update 20170920: Realized that the fix was wrong, the original was correct.
    # Ex_already_seen = []
    # for i_Ex in range(Nbins):
    #     Ex_already_seen.append([])

    # Loop over all transitions and put in the correct pixel:
    for i_tr in range(len(transitions[:,0])):
        Ex = transitions[i_tr,2] - Egs
        # Check if transition is below Ex_max, skip if not
        if Ex < Ex_min or Ex >= Ex_max:
            continue

        # Get bin index for Eg and Ex (initial). Indices are defined with respect to the lower bin edge.
        i_Eg = int(np.floor(transitions[i_tr,6]/bin_width))
        i_Ex = int(np.floor(Ex/bin_width))

        # Read initial spin and parity of level:
        Ji = int(transitions[i_tr,0])
        pi = int(transitions[i_tr,1])
        # Get index for current [Ji,pi] combination in Jpi_list:
        try:
            i_Jpi = Jpi_list.index([Ji,pi])
        except: 
            continue

        # Add B(M1) value and increment count to pixel, respectively
        B_pixel_sum[i_Ex,i_Eg,i_Jpi] += transitions[i_tr,7]

        # 20170920: We thought this was more correct, but now think not.
        # if not Ex in Ex_already_seen[i_Eg]:
        #     B_pixel_count[i_Ex,i_Eg,i_Jpi] += 1
        #     Ex_already_seen[i_Eg].append(Ex)

        B_pixel_count[i_Ex,i_Eg,i_Jpi] += 1 # Original



    # Allocate (Ex,Jpi) matrix to store level density
    rho_ExJpi = np.zeros((Nbins,len(Jpi_list)))
    # Count number of levels for each (Ex, J, pi) pixel.
    for i_l in range(len(levels[:,0])):
        E, J, pi = levels[i_l]
        # Skip if level is outside range:
        if E-Egs >= Ex_max:
            continue
        i_Ex = int(np.floor((E-Egs)/bin_width))
        try:
            i_Jpi = Jpi_list.index([J,pi])
        except:
            continue
        rho_ExJpi[i_Ex,i_Jpi] += 1

    rho_ExJpi /= bin_width # Normalize to bin width, to get density in MeV^-1


    # Calculate gamma strength functions for each Ex, J, pi individually, using the partial level density for each J, pi.
    gSF = np.zeros((Nbins,Nbins,len(Jpi_list)))
    a = prefactor[type] # mu_N^-2 MeV^-2, conversion constant
    for i_Jpi in range(len(Jpi_list)):
        for i_Ex in range(Nbins):
                                         # a *                     <B(M1; Eg, Ex, J, pi)>                                                                             * rho(Ex, J, pi)    
            gSF[i_Ex,:,i_Jpi] = a * div0(B_pixel_sum[i_Ex,:,i_Jpi], B_pixel_count[i_Ex,:,i_Jpi]) * rho_ExJpi[i_Ex, i_Jpi]
            # TODO if the below smoothing is used: Check the form of the level density. When I tried sigma=3 it looked like the gSF dropped a bit, indicating that the level density was lowered. Maybe smoothe it some other way, such as drawing straight lines?
            # gSF[i_Ex,:,i_Jpi] = a * div0(B_pixel_sum[i_Ex,:,i_Jpi], B_pixel_count[i_Ex,:,i_Jpi]) * smoothe(rho_ExJpi[:, i_Jpi], sigma=1)[i_Ex] # Smoothing the initial level density by a gaussian convolution

    # Return the average gSF(Eg) over all (Ex,J,pi)

    return gSF[i_Ex_min:i_Ex_max+1,:,:].mean(axis=(0,2))


def smoothe(array, sigma):
    from scipy.stats import norm
    # Smoothe a one-dimensional array by convoluting with a gaussian of width sigma
    N = len(array)
    index_array = np.linspace(0,N-1, N)
    matrix = np.zeros((N,N))
    for i in range(N):
            try:
                    matrix[i] = array[i]*norm.pdf(index_array, loc=index_array[i], scale=sigma) 
            except IndexError:
                    pass
    # plt.matshow(matrix)
    # plt.show()
    array_smoothed = matrix.sum(axis=0)
    return array_smoothed


def strength_function_average_updated_definition_Jpiaveraging(levels, transitions, Jpi_list, bin_width, Ex_min, Ex_max, type="M1"):
    # 20170815: This function returns the strength function the way we now think is the correct way:
    # By taking only the partial level density corresponding to the specific (Ex, J, pi) pixel in the
    # calculation of the strength function, and then averaging over all three variables to produce
    # <f(Eg)>.
    # 20170821: Another update to this version: To blind ourselves to the same degree as in the 
    # data, we use an average partial level density for each Ex bin to multiply the <B>.
    # This code was first developed in the script strength_function_individual_Jpi.py
    Nbins = int(np.ceil(Ex_max/bin_width)) # Make sure the number of bins cover the whole Ex region.
    # print "Ex_max =", Ex_max, "Nbins*bin_width =", Nbins*bin_width
    bin_array = np.linspace(0,bin_width*Nbins,Nbins+1) # Array of lower bin edge energy values
    bin_array_middle = (bin_array[0:-1]+bin_array[1:])/2 # Array of middle bin values
    # Find index of first and last bin (lower bin edge) where we put counts.
    # It's important to not include the other Ex bins in the averaging later, because they contain zeros which will pull the average down.
    i_Ex_min = int(np.floor(Ex_min/bin_width)) 
    i_Ex_max = int(np.floor(Ex_max/bin_width))    

    Egs = levels[0,0] # Read out the absolute ground state energy, so we can get relative energies later

    # Allocate matrices to store the summed B(M1) values for each pixel, and the number of transitions counted
    B_pixel_sum = np.zeros((Nbins,Nbins))
    B_pixel_count = np.zeros((Nbins,Nbins))

    # Loop over all transitions and put in the correct pixel:
    for i_tr in range(len(transitions[:,0])):
        Ex = transitions[i_tr,2] - Egs
        # Check if transition is below Ex_max, skip if not
        if Ex < Ex_min or Ex >= Ex_max:
            continue

        # Get bin index for Eg and Ex (initial). Indices are defined with respect to the lower bin edge.
        i_Eg = int(np.floor(transitions[i_tr,6]/bin_width))
        i_Ex = int(np.floor(Ex/bin_width))

        # Read initial spin and parity of level:
        Ji = int(transitions[i_tr,0])
        pi = int(transitions[i_tr,1])
        if not [Ji,pi] in Jpi_list:
            continue

        # Add B(M1) value and increment count to pixel, respectively
        B_pixel_sum[i_Ex,i_Eg] += transitions[i_tr,7]
        B_pixel_count[i_Ex,i_Eg] += 1


    # plt.pcolormesh(bin_array_middle, bin_array_middle, B_pixel_sum[:,:,20],norm=LogNorm())
    # plt.show()
    # sys.exit(0)
    # OK -- the B summing seems to work correctly.



    # Allocate Ex array to store level density
    rho_ExJpiavg = np.zeros(Nbins)
    # Count number of levels for each Ex pixel, but only selected (J,pi) values
    for i_l in range(len(levels[:,0])):
        E, J, pi = levels[i_l]
        # Skip if level is outside range:
        if E-Egs >= Ex_max:
            continue
        if not [J,pi] in Jpi_list:
            continue
        i_Ex = int(np.floor((E-Egs)/bin_width))
        rho_ExJpiavg[i_Ex] += 1

    rho_ExJpiavg /= (bin_width*len(Jpi_list)) # Normalize to bin width and number of (J,pi) combinations, to get density in MeV^-1


    # Calculate gamma strength functions for each Ex individually, using the average partial level density
    gSF = np.zeros((Nbins,Nbins))
    a = 11.5473e-9 # mu_N^-2 MeV^-2, conversion constant
    for i_Ex in range(Nbins):
                                     # a *                     <B(M1; Eg, Ex, J, pi)>                                                                             * rho(Ex, J, pi)    
        gSF[i_Ex,:] = a * div0(B_pixel_sum[i_Ex,:], B_pixel_count[i_Ex,:]) * rho_ExJpiavg[i_Ex]

    # Return the average gSF(Eg) over all Ex
    return gSF[i_Ex_min:i_Ex_max+1,:].mean(axis=0)


def level_density_matrix(fname, bin_width=0.2, Emax=12, Ex_low=5, Ex_high=8):
    
    levels = read_energy_levels(fname)

    # Set bin width and range
    Nbins = int(Emax/bin_width)
    Emax_adjusted = bin_width*Nbins
    bins_Ex = np.linspace(0,Emax_adjusted,Nbins+1)
    bins_Ex_middle = (bins_Ex[0:-1]+bins_Ex[1:])/2

    bins_J = np.linspace(0, levels[:,1].max()/2, int(levels[:,1].max()/2)+1)
    print(bins_J)

    Egs = levels[0,0]
    rho_total, tmp = np.histogram(levels[:,0]-Egs, bins=bins_Ex)
    rho_total = rho_total/bin_width # To get density per MeV

    matrix, xedges, yedges = np.histogram2d(levels[:,1]/2, levels[:,0]-Egs, bins=[bins_J,bins_Ex])
    return matrix, xedges, yedges


def level_density_matrix_parity_decomposed(fname, bin_width=0.2, Emax=12, Ex_low=5, Ex_high=8):
    
    levels = read_energy_levels(fname)

    # Set bin width and range
    Nbins = int(Emax/bin_width)
    Emax_adjusted = bin_width*Nbins
    bins_Ex = np.linspace(0,Emax_adjusted,Nbins+1)
    bins_Ex_middle = (bins_Ex[0:-1]+bins_Ex[1:])/2

    bins_J = np.linspace(0, levels[:,1].max()/2, int(levels[:,1].max()/2)+1)
    print(bins_J)

    Egs = levels[0,0]
    # rho_total, tmp = np.histogram(levels[:,0]-Egs, bins=bins_Ex)
    # rho_total = rho_total/bin_width # To get density per MeV

    matrix_plus, xedges, yedges = np.histogram2d(levels[levels[:,2]>0,1]/2, levels[levels[:,2]>0,0]-Egs, bins=[bins_J,bins_Ex])
    matrix_minus, xedges, yedges = np.histogram2d(levels[levels[:,2]<0,1]/2, levels[levels[:,2]<0,0]-Egs, bins=[bins_J,bins_Ex])
    return matrix_plus, matrix_minus, xedges, yedges


def read_interaction_file(filename):
    """
    Read interaction file in KSHELL .snt format
    returns SPE, TBME, msdict, core
    """
    SPE = []
    TBME = []
    core = {}
    msdict = {}
    # Model space dict formatted like
    # msdict = {1: {"n": 0, "l": 3, "j": 7, "Tz": -1},
    #                     2: {"n": 0, "l": 3, "j": 5, "Tz": -1},
    #                     3: {"n": 1, "l": 1, "j": 3, "Tz": -1},
    #                     4: {"n": 1, "l": 1, "j": 1, "Tz": -1},
    #                     5: {"n": 0, "l": 3, "j": 5, "Tz":    1},
    #                     6: {"n": 1, "l": 1, "j": 3, "Tz":    1},
    #                     7: {"n": 1, "l": 1, "j": 1, "Tz":    1},
    #                     8: {"n": 0, "l": 4, "j": 9, "Tz":    1}}
    infile = open(filename, "r")
    while True:
        line = infile.readline()
        if not line:
            break
        if line[0:13] == "! model space":
            line = infile.readline() # Line containing Np, Nn, Z0, A0-Z0
            words = line.split()
            core = {"A":int(words[2])+int(words[3]), "Z":int(words[2])}
            Np = int(words[0])
            Nn = int(words[1])
            # Fill msdict with proton and then neutron orbitals:
            for i in range(Np+Nn):
                line = infile.readline()
                words = line.split()
                msdict[int(words[0])] = {"n":int(words[1]), "l":int(words[2]), "j":int(words[3]), "Tz":int(words[4])}
        if line[0:23] == "!  i  j     <i|H(1b)|j>":
            line = infile.readline()  # Line containing Nspe
            words = line.split()
            Nspe = int(words[0])
            for i in range(Nspe):
                line = infile.readline()
                words = line.split()
                SPE.append(float(words[2]))
            SPE = np.array(SPE)
        if line[0:6] == "! TBME":
            line = infile.readline() # Line containing Ntbme
            words = line.split()
            Ntbme = int(words[0])
            for i in range(Ntbme):
                line = infile.readline()
                words = line.split()
                TBME.append([int(words[0]), int(words[1]), int(words[2]), int(words[3]), int(words[4]), float(words[5])])
            TBME = np.array(TBME) 

    infile.close()

    if len(SPE) < 1 or len(TBME) < 1:
        raise Exception("Something seems to have gone wrong in reading the "
                        "interaction file. "
                        "Please check that the .snt file contains the correct "
                        "formatting, including the lines "
                        "!  i  j     <i|H(1b)|j>"
                        "and "
                        "! TBME")

    return SPE, TBME, msdict, core


def write_interaction_file(filename, SPEs, TBMEs, model_space, core, comments="", mass_scaling=False, scaling_A0=-1, scaling_p=-0.300000):
    outfile = "! interaction file written by write_interaction_file() from statistical.py\n"
    outfile += "! User-added comments, if any:\n"
    outfile += "! " + comments + "\n"

    # Model space formatted like
    #     #            n        l     2j     2tz
    # [[1,         0,     3,     7,    -1],
    #    [2,         0,     3,     5,    -1],
    #    [3,         1,     1,     3,    -1],
    #    [4,         1,     1,     1,    -1],
    #    [5,         0,     3,     5,     1],
    #    [6,         1,     1,     3,     1],
    #    [7,         1,     1,     1,     1],
    #    [8,         0,     4,     9,     1]]
    outfile += "! model space\n" 
    # Write a line with (num proton orbitals)    (num neutron orbitals)    (Z_core)     (N_core)
    outfile += " {:3d} {:3d}    {:3d} {:3d}\n".format(len(np.where(model_space[:,4]<0)[0]), len(np.where(model_space[:,4]>0)[0]), core["Z"], core["A"]-core["Z"])
    for i in range(len(model_space[:,0])):
        outfile += " {:2d}    {:2d}    {:2d}    {:2d}    {:2d}\n".format(model_space[i,0],model_space[i,1],model_space[i,2],model_space[i,3],model_space[i,4])

    # SPEs as one-dim array
    outfile += "! interaction\n"
    outfile += "!  i  j     <i|H(1b)|j>\n"
    outfile += "    {:d}     {:d}\n".format(len(SPEs), 0) # Unsure what the last zero signifies.
    for i in range(len(SPEs)):
        outfile += "{:3d} {:3d} {:16.8f}\n".format(model_space[i,0], model_space[i,0], SPEs[i])

    outfile += "! TBME \n"
    if mass_scaling:
        if scaling_A0 < 0:
            raise Exception("Must specify scaling_A0")
        outfile += "        {:5d}     1    {:d} {:f} \n".format(len(TBMEs[:,0]),scaling_A0, scaling_p)
    else:
        outfile += "         {:5d}     {:3d}\n".format(len(TBMEs[:,0]), 0) # Unsure about last zero
    for i in range(len(TBMEs[:,0])):
        outfile += "{:3d} {:3d} {:3d} {:3d}    {:3d}    {:16.8f}\n".format(int(TBMEs[i,0]),int(TBMEs[i,1]),int(TBMEs[i,2]),int(TBMEs[i,3]),int(TBMEs[i,4]),TBMEs[i,5])


    with open(filename, 'w') as f:
        f.write(outfile)

    return True


def write_interaction_file_msdict(filename, SPEs, TBMEs, model_space, core, comments="", mass_scaling=False, scaling_A0=-1, scaling_p=-0.300000):
    outfile = "! interaction file written by write_interaction_file() from statistical.py\n"
    outfile += "! User-added comments, if any:\n"
    outfile += "! " + comments + "\n"

    # Model space formatted like
    # ms = {1: {"n": 0, "l": 3, "j": 7, "Tz": -1},
    #            2: {"n": 0, "l": 3, "j": 5, "Tz": -1},
    #            3: {"n": 1, "l": 1, "j": 3, "Tz": -1},
    #            4: {"n": 1, "l": 1, "j": 1, "Tz": -1},
    #            5: {"n": 0, "l": 3, "j": 5, "Tz":    1},
    #            6: {"n": 1, "l": 1, "j": 3, "Tz":    1},
    #            7: {"n": 1, "l": 1, "j": 1, "Tz":    1},
    #            8: {"n": 0, "l": 4, "j": 9, "Tz":    1}}

    # Convert it to the format 
    #     #            n        l     2j     2tz
    # [[1,         0,     3,     7,    -1],
    #    [2,         0,     3,     5,    -1],
    #    [3,         1,     1,     3,    -1],
    #    [4,         1,     1,     1,    -1],
    #    [5,         0,     3,     5,     1],
    #    [6,         1,     1,     3,     1],
    #    [7,         1,     1,     1,     1],
    #    [8,         0,     4,     9,     1]]    
    model_space_list = []
    for i in list(model_space.keys()):
        model_space_list.append([i,model_space[i]["n"],model_space[i]["l"],model_space[i]["j"],model_space[i]["Tz"]])
    model_space = np.array(model_space_list)

    outfile += "! model space\n" 
    # Write a line with (num proton orbitals)    (num neutron orbitals)    (Z_core)     (N_core)
    outfile += " {:3d} {:3d}    {:3d} {:3d}\n".format(len(np.where(model_space[:,4]<0)[0]), len(np.where(model_space[:,4]>0)[0]), core["Z"], core["A"]-core["Z"])
    for i in range(len(model_space[:,0])):
        outfile += " {:2d}    {:2d}    {:2d}    {:2d}    {:2d}\n".format(model_space[i,0],model_space[i,1],model_space[i,2],model_space[i,3],model_space[i,4])

    # SPEs as one-dim array
    outfile += "! interaction\n"
    outfile += "!  i  j     <i|H(1b)|j>\n"
    outfile += "    {:d}     {:d}\n".format(len(SPEs), 0) # Unsure what the last zero signifies.
    for i in range(len(SPEs)):
        outfile += "{:3d} {:3d} {:16.8f}\n".format(model_space[i,0], model_space[i,0], SPEs[i])

    outfile += "! TBME \n"
    if mass_scaling:
        if scaling_A0 < 0:
            raise Exception("Must specify scaling_A0")
        outfile += "        {:5d}     1    {:d} {:f} \n".format(len(TBMEs[:,0]),scaling_A0, scaling_p)
    else:
        outfile += "         {:5d}     {:3d}\n".format(len(TBMEs[:,0]), 0) 
    for i in range(len(TBMEs[:,0])):
        outfile += "{:3d} {:3d} {:3d} {:3d}    {:3d}    {:16.8f}\n".format(int(TBMEs[i,0]),int(TBMEs[i,1]),int(TBMEs[i,2]),int(TBMEs[i,3]),int(TBMEs[i,4]),TBMEs[i,5])


    with open(filename, 'w') as f:
        f.write(outfile)

    return True


def write_interaction_file_msdict(filename, SPEs, TBMEs, model_space, core, comments="", mass_scaling=False, scaling_A0=-1, scaling_p=-0.300000):
    outfile = "! interaction file written by write_interaction_file() from statistical.py\n"
    outfile += "! User-added comments, if any:\n"
    outfile += "! " + comments + "\n"

    # Model space formatted like
    # ms = {1: {"n": 0, "l": 3, "j": 7, "Tz": -1},
    #            2: {"n": 0, "l": 3, "j": 5, "Tz": -1},
    #            3: {"n": 1, "l": 1, "j": 3, "Tz": -1},
    #            4: {"n": 1, "l": 1, "j": 1, "Tz": -1},
    #            5: {"n": 0, "l": 3, "j": 5, "Tz":    1},
    #            6: {"n": 1, "l": 1, "j": 3, "Tz":    1},
    #            7: {"n": 1, "l": 1, "j": 1, "Tz":    1},
    #            8: {"n": 0, "l": 4, "j": 9, "Tz":    1}}

    # Convert it to the format 
    #     #            n        l     2j     2tz
    # [[1,         0,     3,     7,    -1],
    #    [2,         0,     3,     5,    -1],
    #    [3,         1,     1,     3,    -1],
    #    [4,         1,     1,     1,    -1],
    #    [5,         0,     3,     5,     1],
    #    [6,         1,     1,     3,     1],
    #    [7,         1,     1,     1,     1],
    #    [8,         0,     4,     9,     1]]    
    model_space_list = []
    for i in list(model_space.keys()):
        model_space_list.append([i,model_space[i]["n"],model_space[i]["l"],model_space[i]["j"],model_space[i]["Tz"]])
    model_space = np.array(model_space_list)

    outfile += "! model space\n" 
    # Write a line with (num proton orbitals)    (num neutron orbitals)    (Z_core)     (N_core)
    outfile += " {:3d} {:3d}    {:3d} {:3d}\n".format(len(np.where(model_space[:,4]<0)[0]), len(np.where(model_space[:,4]>0)[0]), core["Z"], core["A"]-core["Z"])
    for i in range(len(model_space[:,0])):
        outfile += " {:2d}    {:2d}    {:2d}    {:2d}    {:2d}\n".format(model_space[i,0],model_space[i,1],model_space[i,2],model_space[i,3],model_space[i,4])

    # SPEs as one-dim array
    outfile += "! interaction\n"
    outfile += "!  i  j     <i|H(1b)|j>\n"
    outfile += "    {:d}     {:d}\n".format(len(SPEs), 0) # Unsure what the last zero signifies.
    for i in range(len(SPEs)):
        outfile += "{:3d} {:3d} {:16.8f}\n".format(model_space[i,0], model_space[i,0], SPEs[i])

    outfile += "! TBME \n"
    if mass_scaling:
        if scaling_A0 < 0:
            raise Exception("Must specify scaling_A0")
        outfile += "        {:5d}     1    {:d} {:f} \n".format(len(TBMEs[:,0]),scaling_A0, scaling_p)
    else:
        outfile += "         {:5d}     {:3d}\n".format(len(TBMEs[:,0]), 0) 
    for i in range(len(TBMEs[:,0])):
        outfile += "{:3d} {:3d} {:3d} {:3d}    {:3d}    {:16.8f}\n".format(int(TBMEs[i,0]),int(TBMEs[i,1]),int(TBMEs[i,2]),int(TBMEs[i,3]),int(TBMEs[i,4]),TBMEs[i,5])


    with open(filename, 'w') as f:
        f.write(outfile)

    return True


def spider(fnames, names, type="M1", threshold=0.1, spinwindow=[], Eg_low=0, Eg_high=1e9, Ex_low=0, Ex_high=1e9, scale=2):

    Nsp = np.ceil(np.sqrt(len(fnames))).astype(int)
    f, ax_list = plt.subplots(Nsp,Nsp,squeeze=False,sharex='col', sharey='row')

    for i in range(len(fnames)):
        fname = fnames[i]
        name = names[i]
        ax = ax_list[i%Nsp][int((i-i%Nsp)/Nsp)]


        levels = read_energy_levels(fname)
        Egs = levels[0,0]

        Ex_high = min(Ex_high, levels[:,0].max()-Egs)
        Eg_high = min(Eg_high, levels[:,0].max()-Egs)
    
    
        levels_plot_J = []
        levels_plot_Ex = []
        for iEx in range(len(levels[:,0])):
            # print levels[iEx,:]
            J2 = levels[iEx,1]
            par = levels[iEx,2]
            if len(spinwindow) > 0 and not ([J2,par] in spinwindow or [J2-2,par] in spinwindow or [J2+2,par] in spinwindow):
                continue
            Ex = levels[iEx,0]-Egs
            if Ex < Ex_low or Ex > Ex_high:
                continue
    
            levels_plot_J.append(J2/2)
            levels_plot_Ex.append(Ex)
    
        ax.plot(levels_plot_J, levels_plot_Ex, 'o', color='grey', linewidth=0.5)
        ax.set_xlim([levels[:,1].min()/2-1,levels[:,1].max()/2+1])
        ax.set_title(name+r'$\,E_\gamma\in[{:.1f},{:.1f}]$'.format(Eg_low,Eg_high))
        ax.set_ylabel(r'$E_x\,\mathrm{[MeV]}$')
        ax.set_xlabel(r'$J\,\mathrm{[\hbar]}$')
    
    
    
        transitions = read_transition_strengths(fname, type=type)
        for iEx in range(len(transitions[:,0])):
            J2i = int(transitions[iEx,0])
            pari = int(transitions[iEx,1])
            if len(spinwindow) > 0 and not [J2i,pari] in spinwindow:
                continue
            B = transitions[iEx,7]
            Eg = transitions[iEx,6]
            if B < threshold or Eg<Eg_low or Eg>Eg_high:
                continue
            Ei = transitions[iEx,2]-Egs
            if Ei < Ex_low or Ei > Ex_high:
                continue
            J2f = int(transitions[iEx,3])
            parf = int(transitions[iEx,4])
            Ef = transitions[iEx,5]-Egs
            ax.plot([J2i/2,J2f/2],[Ei,Ef], color='teal', linewidth=(scale*B))
    
    return f, ax_list















from setuptools import setup

setup(
    name = 'kshell-utilities',
    version = '1.0.1.3',    
    description = 'Handy utilities for handling nuclear shell model calculations from KSHELL',
    url = 'https://github.com/GaffaSnobb/kshell-utilities',
    author = 'Jon Kristian Dahl',
    author_email = 'jonkd@uio.no',
    packages = ['kshell_utilities', 'tests'],
    install_requires = ['numpy', 'matplotlib'],

    classifiers = [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering :: Physics',
    ],
)